/* Copyright 2004, 2005, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Client-handling code for the THINC sound daemon (tsd.c).
 *
 * Authors: 
 * Leo Kim - Network Computing Lab - Columbia University
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */
/* set tabstop = 8 */

#include <stdio.h>
#include <stdlib.h>

/* for errno */
#include <errno.h>
#include <string.h>

#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

#include <sys/socket.h>

#include "tsd.h"

/* Local funcs */
static int tsd_handleClientConnection(struct tsd_state *state);
static int tsd_initClient(struct tsd_state *state, int clientSock);
static int tsd_readSoundData(struct tsd_client *client, struct tsd_state *state);
static int tsd_handleTCPActivity(struct tsd_state *state,
	struct tsd_client *client);
static int tsd_handleUDPActivity(struct tsd_state *state,
	struct tsd_client *client);

/* handles new connections and scans list of clients for activity */
void tsd_handleNetworkActivity(struct tsd_state *state)
{
	struct tsd_client *client = state->clients;
	int err;

	/* first check connected clients */
	while (client) {
	        if(state->udp && (FD_ISSET(client->udp_sock, &state->rfds)) ) {
			err = tsd_handleUDPActivity(state, client);
			if(err == T_ERROR) {
				tsd_closeClient(client);
				goto next;
			}
		}
	        if (FD_ISSET(client->tcp_sock, &state->rfds)) {
			err = tsd_handleTCPActivity(state, client);
			if(err == T_ERROR) {
				tsd_closeClient(client);
				goto next;
			}
		}
next:
	        client = client->next;
	}

	/* now take care of any new connections */
	if (FD_ISSET(state->sock, &state->rfds)) {
		tsd_handleClientConnection(state);
	}
}

static int tsd_handleTCPActivity(struct tsd_state *state,
	struct tsd_client *client)
{
	int err;
	thinc_header h;

	err = state->read(client, &h, 0, THINC_HEADERSIZE);
	if(err == T_ERROR) {
		fprintf(stderr, "Client died\n");
		return T_ERROR;
	}
	switch(h.type) {
		case T_SND_DATA:
			err = tsd_readSoundData(client, state);
			break;
		default:
			fprintf(stderr, "Received unknown "
				"command %d from client\n",
				h.type);
			return T_ERROR;
	}
	if(err == T_ERROR) {
		fprintf(stderr, "Couldn't handle client "
			"command %d\n", h.type);
		return T_ERROR;
	}

	return T_OK;
}

/* for now same as readSoundData since that's all the data we can receive
 * on the UDP socket */
static int tsd_handleUDPActivity(struct tsd_state *state,
	struct tsd_client *client)
{
	int err;
	thinc_sndUDPData *data = state->udp_data;
	unsigned int size, space;

	err = read(client->udp_sock, data, T_SND_UDP_DATA_SIZE);
	if(err == -1) {
		fprintf(stderr, "Error: Couldn't read UDP data from "
			"client: %s\n", strerror(errno));
		return T_ERROR;
	}

	/* for now we ignore id */
	size = ntohl(data->size);
	if (size > state->bufsize) {
		fprintf(stderr, "Client sent more data (%d) than we are allowed "
			"to read (%d)\n", size, state->bufsize);
		return T_ERROR;
	}

	/* see if we have enough space in the buffer for the whole thing. note
	 * that we blindly overwrite data. and that's ok. */
	space = state->capture_tail - state->capture_head;
	if (space < size) {
		memcpy(state->capture_head, data->data, space);
		memcpy(state->capture, data->data + space, size - space);
		state->capture_head = state->capture + (size-space);
	} else {
		memcpy(state->capture_head, data->data, size);
		state->capture_head += size;
	}
	if(state->capture_head == state->capture_tail)
		state->capture_head = state->capture;

	/* now tell the driver how much we read */
	/* FIXME: Should have some command to pass this info! */
	err = write(state->mmapfd, &size, sizeof(size));
	if(err == T_ERROR)
		return T_ERROR;

	return T_OK;
}

/* This function returns T_OK if a client connection has been successfully  
 * added to the client list.
 *
 * T_ERROR is returned if there is a problem with the socket, or if there is
 * no more memory to support a new client structure.
 */
static int tsd_handleClientConnection(struct tsd_state *state)
{
	const char *func = __func__;
	int sock, opt;
	struct sockaddr_in addr_c;
	socklen_t addrlen;

	fprintf(stdout, "Detecting connection attempt.\n");

	addrlen = sizeof(addr_c);
	sock = accept(state->sock, (struct sockaddr *) &addr_c,
			&addrlen);
	if (sock == -1) {
		fprintf(stderr, "%s: Couldn't accept() new client. %s.\n", 
				func, strerror(errno));
		fprintf(stderr, "Address was %s\n", inet_ntoa(addr_c.sin_addr));
		return T_ERROR;
	}

	/* disable nagle's algorithm. Non fatal if we can't do it */
	opt = 1;
	if(setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char*)&opt, 
		sizeof(opt)) < 0) {
		fprintf(stderr, "Warning: Couldn't disable nagle's "
			"algorithm. "
			"Your performance may suffer\n");
	}

	if (tsd_initClient(state, sock) == T_ERROR) {
		fprintf(stderr, "Error occurred while handling new "
				"connection.\n");
		return T_ERROR;
	}
	
	fprintf(stdout, "Got connection from %s\n", inet_ntoa(addr_c.sin_addr));
	return T_OK;
}

void tsd_closeClient(struct tsd_client *client)
{
	const char *func = __func__;
	struct tsd_state *state = client->state;
	struct tsd_client *prev, *curr;

	fprintf(stdout, "%s: Got client disconnect.\n", func);
	if (close(client->tcp_sock) < 0) {
		fprintf(stderr, "%s: %s\n", func, strerror(errno));
	}

	prev = curr = state->clients;

	while (curr) {
		if (curr != client) {
			prev = curr;
			curr = curr->next;
			continue;
		}
		/* found it. */
		if (prev == curr) {
			state->clients = curr->next;
		} else {
			prev->next = curr->next;
		}
		fprintf(stdout, "%s: Removing client\n", func);
		free(curr);
		return;
	}

	/* maybe we didn't find it???? */
	fprintf(stdout, "%s: Extremely strange error. Could not find client "
			"structure to delete.\n", func);
}

/* This function is responsible for initializing new client connections and
 * adding the client to the main client list. */
static int tsd_initClient(struct tsd_state *state, int clientSock)
{
	const char *func = __func__;
	int err;
	struct tsd_client *newClient;
	thinc_sndInit si;

	newClient = (struct tsd_client *)malloc(sizeof(struct tsd_client));
	if (!newClient) {
		/* arggh! */
		fprintf(stderr, "Could not allocate memory for a new "
				"client.\n");
		close(clientSock);
		return T_ERROR;
	}

	/* Do the security thing */
	if(state->beSecure) {
		newClient->ssl = NULL;
		fprintf(stdout, "Establishing secure connection with "
				"client...\n");
		err = thinc_ssl_accept(clientSock, state->ctx, &newClient->ssl);
		if (err == T_SSL_CLOSED || err == T_SSL_ERROR ) {
			fprintf(stderr, "ERROR: %s: Couldn't initialize secure "
					"connection with new client: %d\n", 
					func, err);
			free(newClient);
			close(clientSock);
			return T_ERROR;
		} else if (err == T_SSL_WOULDBLOCK) {
			/* didn't finish, but perhaps it will */
			fprintf(stderr, "%s: SSL handshake didn't finish\n", 
					func);
		}
		newClient->isSecure = T_TRUE;
		fprintf(stdout, "Secure connection established.\n");
	} else {
		newClient->isSecure = T_FALSE;
		newClient->ssl = NULL;
	}

	if (fcntl(clientSock, F_SETFL, O_NONBLOCK) < 0) {
		fprintf(stderr, "%s: Couldn't set non-blocking "
				"flag. %s.\n", func, strerror(errno));
		free(newClient);
		close(clientSock);
		return T_ERROR;
	}

	newClient->state = state;
	newClient->tcp_sock = clientSock;
	/* Set the stage. har har. */
	newClient->stage = TSD_C_STAGE_NEW;
	newClient->next = state->clients;

	/* Add to head. from now on, call tsd_closeClient if something goes
	 * bad */
	state->clients = newClient;

	/* send init message */
	if(state->udp) {
		if(tsd_initClientUDP(newClient, state) == T_ERROR) {
			fprintf(stderr, "Error: %s: Couldn't initialize UDP "
				"connection with client\n", func);
			tsd_closeClient(newClient);
			return T_ERROR;
		}
		fprintf(stdout, "Info: Using UDP port %d\n",
			newClient->udp_port);
		si.port = newClient->udp_port;
	} else {
		si.port = 0;
		newClient->udp_sock = -1;
	}
	si.port = htons(si.port);
	/* :BUG: We could potentially block on writing to this client 
	 * while it initializes, thus blocking other clients from 
	 * playing back. */
	if(state->write(newClient, &si, 0, T_SND_INIT_SIZE) == T_ERROR) {
		fprintf(stderr, "Error: %s: Couldn't send sound init message "
			"to new client\n", func);
		tsd_closeClient(newClient);
		return T_ERROR;
	}
	/* TODO: if wanted to traverse client firewall, have it send us
	 * a UDP packet. note that the socket has to be non-blocking for
	 * this to work :( */
#if 0
	{
		int dummy;
		read(newClient->udp_sock, &dummy, sizeof(dummy)), errno);
	}
#endif

	/* If we're playing something, send the parameters of the current 
	 * stream. */
	/* TODO: Do the same for capture */
	if (tsd_isOpen(state, T_MODE_PLAYBACK)) {
		fprintf(stdout, "Sound playback detected. Sending playback "
				"parameters to client.\n");
		int written;
		thinc_header *header = (thinc_header *)state->cmdbuf;
		header->type = T_SND_PLAYBACK_OPEN;
		memcpy(state->cmdbuf + THINC_HEADERSIZE, &state->openParams, 
				T_SND_OPEN_SIZE);

		/* :BUG: We could potentially block on writing to this client 
		 * while it initializes, thus blocking other clients from 
		 * playing back. */
		written = state->write(newClient, state->cmdbuf, 0, 
				THINC_HEADERSIZE + T_SND_OPEN_SIZE);
		if (written == T_ERROR) {
			fprintf(stderr, "%s: Problem writing to client: %s\n", 
					func, strerror(errno));
			tsd_closeClient(newClient);
			return T_ERROR;
		}
	}

	fprintf(stdout, "Info: Added new client\n");
	return T_OK;
} 

static int tsd_readSoundData(struct tsd_client *client, struct tsd_state *state)
{
	int err;
	thinc_sndData data;
	unsigned int size, space;

	err = state->read(client, &data, 0, T_SND_DATA_SIZE);
	if(err == T_ERROR) {
		return T_ERROR;
	}

	/* for now we ignore id */
	size = ntohl(data.size);
	if (size > state->bufsize) {
		fprintf(stderr, "Client sent more data (%d) than we are allowed "
			"to read (%d)\n", size, state->bufsize);
		return T_ERROR;
	}

	/* see if we have enough space in the buffer for the whole thing. note
	 * that we blindly overwrite data. and that's ok. */
	space = state->capture_tail - state->capture_head;
	if (space < size) {
		err = state->read(client, state->capture_head, 0, space);
		if(err == T_ERROR)
			return T_ERROR;
		err = state->read(client, state->capture, 0, size - space);
		if(err == T_ERROR)
			return T_ERROR;
		state->capture_head = state->capture + (size-space);
	} else {
		err = state->read(client, state->capture_head, 0, size);
		if(err == T_ERROR)
			return T_ERROR;
		state->capture_head += size;
	}
	if(state->capture_head == state->capture_tail)
		state->capture_head = state->capture;

	/* now tell the driver how much we read */
	/* FIXME: Should have some command to pass this info! */
	err = write(state->mmapfd, &size, sizeof(size));
	if(err == T_ERROR)
		return T_ERROR;

	return T_OK;
}
