/* Copyright 2004, 2005, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Code for handling driver communication for the THINC sound daemon.
 *
 * Authors: 
 * Leo Kim - Network Computing Lab - Columbia University
 * Rebecca Collins - Network Computing Lab - Columbia University
 * Ricardo Baratto - Network Computing Lab - Columbia University
 */
/* set tabstop = 8 */

#include <stdio.h>

/* for errno */
#include <errno.h>
#include <string.h>

#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>

/* for hton* */
#include <netinet/in.h>

#include "tsd.h"

/* prototypes */
static int tsd_handleDriverCtrlCmd(struct tsd_state *state);
static int tsd_handleDriverPlay(struct tsd_state *state);
static int tsd_handleDriverOpen(struct tsd_state *state, 
		struct thinc_snd_cmd *cmd, int which);
static int tsd_handleDriverClose(struct tsd_state *state, 
		struct thinc_snd_cmd *cmd, int which);
static int tsd_handleDriverPlayOpen(struct tsd_state *state, 
		struct thinc_snd_cmd *cmd);
static int tsd_handleDriverPlayEnd(struct tsd_state *state, 
		struct thinc_snd_cmd *cmd);
static int tsd_handleDriverCaptureOpen(struct tsd_state *state, 
		struct thinc_snd_cmd *cmd);
static int tsd_handleDriverCaptureEnd(struct tsd_state *state, 
		struct thinc_snd_cmd *cmd);

static int tsd_handleDriverVolumeDown(struct tsd_state *state, 
		struct thinc_snd_cmd *cmd);
static int tsd_handleDriverVolumeUp(struct tsd_state *state, 
		struct thinc_snd_cmd *cmd);
static unsigned int tsd_driverDeletePlayCmd(int fd);
static unsigned int tsd_driverGenerateOpenCmd(int fd);
static unsigned int tsd_driverGetNumPlayCmds(int fd);
static unsigned int tsd_driverGetPlayCmdOffset(int fd);
unsigned int tsd_driverResetPlayCmds(int fd);

static int tsd_playTCP(struct tsd_state *state, unsigned int offset);
static int tsd_playUDP(struct tsd_state *state, unsigned int offset);

/* array of function pointers to handle commands from driver */
static int (*t_snd_cmd_func[T_SND_NUM_CMDS]) (struct tsd_state *state, 
		struct thinc_snd_cmd *cmd); 
/* function pointer for play command */
static int (*t_snd_play_func) (struct tsd_state *state);

void tsd_initDriverCmdFuncs(void)
{
	t_snd_play_func = tsd_handleDriverPlay;
	t_snd_cmd_func[T_SND_CMD_PLAYBACK_OPEN] = tsd_handleDriverPlayOpen;
	t_snd_cmd_func[T_SND_CMD_PLAYBACK_END] = tsd_handleDriverPlayEnd;
	t_snd_cmd_func[T_SND_CMD_VOLUME_DOWN] = tsd_handleDriverVolumeDown;
	t_snd_cmd_func[T_SND_CMD_VOLUME_UP] = tsd_handleDriverVolumeUp;
	t_snd_cmd_func[T_SND_CMD_CAPTURE_OPEN] = tsd_handleDriverCaptureOpen;
	t_snd_cmd_func[T_SND_CMD_CAPTURE_END] = tsd_handleDriverCaptureEnd;
}

/* called by tsd_waitForActivity to handle any wakeups from the driver */
void tsd_handleDriverActivity(struct tsd_state *state) 
{
	const char *func = __func__;
	int res;

	if (!FD_ISSET(state->mmapfd, &state->rfds)) {
		/* Nothing to do here */
		return;
	}


	/* check for any data to play first */
	res = t_snd_play_func(state);
	if (res == T_ERROR) {
		/* uh oh */
		fprintf(stderr, "%s: An error occurred while attempting to "
			"send data for playback to client.\n", func);
		return;
	}

	/* check for commands */
	res = tsd_handleDriverCtrlCmd(state);
	if (res == T_ERROR) {
		/* uh oh */
		fprintf(stderr, "%s: An error occurred while attempting to send "
			"command data for playback to client.\n", func);
	}

#if 0
	if (state->fds[state->mmapPoll].revents & POLLPRI) { 
		res = tsd_handleDriverCtrlCmd(state);
		if (res == T_ERROR) {
			/* uh oh */
			fprintf(stderr, "%s: An error occurred while attempting to send "
					"command data for playback to client.\n", func);
		}
	}
#endif
}

/* Handles auxiliary soundcard functionality like volume control */
static int tsd_handleDriverCtrlCmd(struct tsd_state *state)
{
	struct thinc_snd_cmd cmd;
	int retval;

	retval = read(state->mmapfd, &cmd, sizeof(struct thinc_snd_cmd));
	if (retval == 0) {
		/* The ctrl command queue was empty. */
		return T_OK;
	} else if (retval < 0) {
		/* Driver had problem copying to user. But why? */
		fprintf(stderr, "Problem reading driver command.\n");
		return T_ERROR;
	}

	/* Dispatch to helper function. */
	if (t_snd_cmd_func[cmd.id](state, &cmd) == T_ERROR) {
		fprintf(stderr, "Error while handling command %d\n", cmd.id);
		return T_ERROR;
	}

	return T_OK;
}

static int tsd_handleDriverPlay(struct tsd_state *state) 
{
	unsigned int numcmds, offset;
	int written;

	if ( (numcmds = tsd_driverGetNumPlayCmds(state->mmapfd)) == 0) {
		/* Nothing to do here. */
		return T_OK;
	}

	/* send to clients only if there are actually clients out there. */
	if (!state->clients) {
		tsd_driverDeletePlayCmd(state->mmapfd);
		return T_OK;
	}

	/* :HACK: If for some screwed up reason the application wants to send 
	 * sound to the client and has not opened the sound card, or if tsd 
	 * was killed and restarted and does not realize that an application
	 * has already opened the sound card, then we try to retrieve the 
	 * last known setting by querying the driver for the last known 
	 * parameters. This might not always work, but then again, this is 
	 * broken behavior. */
	if (tsd_isOpen(state, T_MODE_PLAYBACK) == T_FALSE) {
		fprintf(stdout, "Warning: Open sound parameters not properly "
				"set on tsd. Querying driver for proper "
				"parameters.\n");
		/* generate the open parameters from the driver */
		if (tsd_driverGenerateOpenCmd(state->mmapfd) < 0) {
			fprintf(stderr, "Serious error: Could not "
					"generate open command from "
					"driver.\n");
			return T_ERROR;
		}
		if (tsd_handleDriverCtrlCmd(state) == T_ERROR) {
			return T_ERROR;
		}
	}

	/* Get the offset from which to read from the shared buffer. */
	offset = tsd_driverGetPlayCmdOffset(state->mmapfd);

	if(state->udp) {
		written = tsd_playUDP(state, offset);
	} else {
		written = tsd_playTCP(state, offset);
	}

	/* Decrement command count. */
	tsd_driverDeletePlayCmd(state->mmapfd);

	return written;
}

static int tsd_playTCP(struct tsd_state *state, unsigned int offset)
{
	const char *func = __func__;
	struct thinc_snd_cmd *cmd;
	struct tsd_client *client = state->clients;
	thinc_header *header;
	thinc_sndData *sndData;
	int written;

	cmd = (struct thinc_snd_cmd *)(state->play + offset);

	/* Create header from command buf. */
	header = (thinc_header *)state->cmdbuf;
	sndData = (thinc_sndData *)(state->cmdbuf + THINC_HEADERSIZE);
	header->type = T_SND_DATA;
	sndData->timestamp = htonl(cmd->type.play.timestamp);

	/* Just set the ID to 0 now. When we support mixing, this 
	 * will become significant. */
	sndData->id = 0;
	sndData->size = htonl(cmd->type.play.length);

	while (client) {
		/* send header */
		written = state->write(client, state->cmdbuf, 0, 
			THINC_HEADERSIZE + T_SND_DATA_SIZE);
		if (written == T_ERROR) {
			fprintf(stderr, "%s: Problem writing header to "
					"client: %s\n", func, strerror(errno));
			/* close the client */
			tsd_closeClient(client);
			return written;
		}
		/* send shared data */
		written = state->write(client, state->play, 
			cmd->type.play.offset, cmd->type.play.length);
		if (written == T_ERROR) {
			fprintf(stderr, "%s: Problem writing to client: %s.\n", 
					func, strerror(errno));
			/* close the client */
			tsd_closeClient(client);
			return written;
		}

		client->count += written;
		client = client->next;
	}

#if 0
	fprintf(stdout, "%s: Wrote %d bytes to clients.\n", func, toWrite);
#endif
	return T_OK;
}

static int tsd_playUDP(struct tsd_state *state, unsigned int offset)
{
	struct thinc_snd_cmd *cmd;
	int err, left;
	thinc_sndUDPData *data = state->udp_data;
	struct tsd_client *client;
	char *src;

	cmd = (struct thinc_snd_cmd *)(state->play + offset);

	data->header.type = T_SND_DATA;
	data->timestamp = htonl(cmd->type.play.timestamp);
	data->id = 0;

	left = cmd->type.play.length;

	src = state->play + cmd->type.play.offset;
	while(left >= T_SND_UDP_PKT_SIZE) {
		data->size = htonl(T_SND_UDP_PKT_SIZE);

		memcpy(data->data, src, T_SND_UDP_PKT_SIZE);
		client = state->clients;
		while(client) {
			err = write(client->udp_sock, data, T_SND_UDP_DATA_SIZE);
			if(err == -1) {
				fprintf(stderr, "Error: Couldn't write UDP "
					"data to client: %s\n",
					strerror(errno));
				tsd_closeClient(client);
				return T_ERROR;
			}
			client->count += err;
			client = client->next;
		}
		left -= T_SND_UDP_PKT_SIZE;
		src += T_SND_UDP_PKT_SIZE;
	}
	if(left) {
		data->size = htonl(left);
		memcpy(data->data, src, left);
		client = state->clients;
		/* account for header size */
		left += T_SND_UDP_DATA_HSIZE;
		while(client) {
			err = write(client->udp_sock, data, left);
			if(err == -1) {
				fprintf(stderr, "Error: Couldn't write UDP "
					"data to client: %s\n",
					strerror(errno));
				tsd_closeClient(client);
				return T_ERROR;
			}
			client->count += err;
			client = client->next;
		}
	}

	return T_OK;
}

static int tsd_handleDriverPlayOpen(struct tsd_state *state, 
	struct thinc_snd_cmd *cmd)
{
	return tsd_handleDriverOpen(state, cmd, T_MODE_PLAYBACK);
}

static int tsd_handleDriverPlayEnd(struct tsd_state *state, 
	struct thinc_snd_cmd *cmd)
{
	return tsd_handleDriverClose(state, cmd, T_MODE_PLAYBACK);
}

static int tsd_handleDriverCaptureOpen(struct tsd_state *state, 
	struct thinc_snd_cmd *cmd)
{
	return tsd_handleDriverOpen(state, cmd, T_MODE_CAPTURE);
}

static int tsd_handleDriverCaptureEnd(struct tsd_state *state, 
	struct thinc_snd_cmd *cmd)
{
	return tsd_handleDriverClose(state, cmd, T_MODE_CAPTURE);
}

static int tsd_handleDriverOpen(struct tsd_state *state, 
		struct thinc_snd_cmd *cmd, int which)
{
#if 1
	const char *func = __func__;
#endif
	int res;
	thinc_header *header;
	thinc_sndOpen *open;

#if 0
	fprintf(stdout, "%s: Got open message from driver.\n", func); 
#endif

	header = (thinc_header *)state->cmdbuf;
	if(which == T_MODE_PLAYBACK) {
		header->type = T_SND_PLAYBACK_OPEN;
		/* set open flag */
		state->playIsOpen = T_TRUE;
	} else {
		header->type = T_SND_CAPTURE_OPEN;
		state->captureIsOpen = T_TRUE;
	}

	open = (thinc_sndOpen *)(state->cmdbuf + THINC_HEADERSIZE);
	open->id = 0;
	open->bits = cmd->type.open.bits;
	open->rate = htons(cmd->type.open.rate);
	open->channels = cmd->type.open.channels;
	open->endian = cmd->type.open.endian;

	/* Set as the parameters of the current stream. We use this 
	 * to send to the client if playback is in progress. (See 
	 * tsd_initClient.)  */
	memcpy(&state->openParams, open, T_SND_OPEN_SIZE);

#if 0
	fprintf(stdout, "%s: header is type %d, length %d\n", func, 
		header->type, header->length);
	fprintf(stdout, "%s: open is %d bits, %d channels, %d endian, %d rate\n", 
			func, open->bits, open->channels, open->endian, 
			cmd->type.open.rate);
	fprintf(stdout, "%s: setting client playback at %d bits, %d channels, " 
			"%s endian, %d rate\n", func, state->openParams.bits, 
			state->openParams.channels, 
			(state->openParams.endian == T_BIG) ? "big" : "little", 
			cmd->type.open.rate);
#endif
	res = tsd_writeToClients(state, state->cmdbuf, 
			THINC_HEADERSIZE + T_SND_OPEN_SIZE);
	fprintf(stdout, "%s: wrote open message to clients.\n", func);
	if (res != T_OK) {
		fprintf(stderr, "%s: An error occurred while writing the open "
				"command to clients.\n", func);
		return T_ERROR;
	}

	return T_OK; 
}

static int tsd_handleDriverClose(struct tsd_state *state, 
	struct thinc_snd_cmd *cmd, int which)
{
	const char *func = __func__;
	int res;
	thinc_header *header;
	thinc_sndClose *close;
#if 0
	fprintf(stdout, "%s: Got close message from driver.\n", func); 
#endif

	/* If the sound card happens to already be closed, don't even bother 
	 * sending anything. This happens sometimes when ill-behaved audio apps
	 * like Flash want to keep opening and closing the damn soundcard. */
#if 0
	if (tsd_isOpen(state, T_MODE_PLAYBACK)) {
		return T_OK;
	}
#endif
	header = (thinc_header *)state->cmdbuf;
	if(which == T_MODE_PLAYBACK) {
		fprintf(stdout, "Closing playback\n");
		/* reset open flag */
		state->playIsOpen = T_FALSE;
		header->type = T_SND_PLAYBACK_END;
	} else {
		fprintf(stdout, "Closing capture\n");
		state->captureIsOpen = T_FALSE;
		header->type = T_SND_CAPTURE_END;
	}

	close = (thinc_sndClose *)(state->cmdbuf + THINC_HEADERSIZE);
	/* Only using ID 0 now. When we support mixing, this 
	 * will become significant. */
	close->id = 0;

	/* send the close message */
	res = tsd_writeToClients(state, state->cmdbuf, 
		THINC_HEADERSIZE + T_SND_CLOSE_SIZE);
#if 0
	fprintf(stdout, "%s: Sent close message.\n", func); 
#endif
	if (res != T_OK) {
		fprintf(stderr, "%s: An error occurred while writing the close "
				"command to clients.\n", func);
		return T_ERROR;
	}

	return T_OK; 
}

/* Empty for now. */
static int tsd_handleDriverVolumeDown(struct tsd_state *state, 
		struct thinc_snd_cmd *cmd)
{
	const char *func = __func__;

	fprintf(stdout, "%s: Got volume up from driver.\n", func); 

	return T_OK;
}

/* Empty for now. */
static int tsd_handleDriverVolumeUp(struct tsd_state *state, 
		struct thinc_snd_cmd *cmd)
{
	const char *func = __func__;

	fprintf(stdout, "%s: Got volume down from driver.\n", func); 

	return T_OK;
}

inline int tsd_isOpen(struct tsd_state *state, int which) 
{
	return which == T_MODE_PLAYBACK ? 
		state->playIsOpen : 
		state->captureIsOpen;
}

/* Wrappers for ioctl calls */
static unsigned int tsd_driverDeletePlayCmd(int fd)
{
	return ioctl(fd, T_SND_IOC_DEL_PLAY_CMD);
}

static unsigned int tsd_driverGenerateOpenCmd(int fd)
{
	return ioctl(fd, T_SND_IOC_OPEN_CTRL_CMD);
}

static unsigned int tsd_driverGetPlayCmdOffset(int fd) 
{
	return ioctl(fd, T_SND_IOC_PLAY_CMD_OFFSET);
}

static unsigned int tsd_driverGetNumPlayCmds(int fd) 
{
	return ioctl(fd, T_SND_IOC_NUM_PLAY_CMDS);
}

unsigned int tsd_driverResetPlayCmds(int fd)
{
	return ioctl(fd, T_SND_IOC_RESET_PLAY_CMDS);
}


