/* Copyright 2007 Ricardo Baratto */
/* Copyright 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * daemon.c: Contains main function and event loop for THINC sound daemon 
 *
 * Authors: 
 * Ricardo Baratto - Network Computing Lab - Columbia University
 */
/* set tabstop = 8 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <signal.h>
#include <getopt.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <signal.h>
#include <poll.h>
#include <sys/types.h>
#include <sys/shm.h>

#include "common.h"
#include "thinc_list.h"
#include "thinc_ssl.h"

#ifdef DEBUG
static int verbose = 1;
#else
static int verbose = 0;
#endif

#define s_verb(...) 			\
	if(verbose) {			\
		s_msg(">> " __VA_ARGS__);	\
	}

struct _s_head;

typedef struct {
	int		net_fd;		/* network listen socket */
	int		unix_fd;	/* local listen socket */

	fd_set		fds;
	int		fdmax;

	thinc_list	apps;
	thinc_list	clients;

	/* for SSL */
	SSL_CTX*	ctx;

	char*		un_path;
	unsigned int 	un_path_len;

	/* some flags */
	unsigned int	secure:1;
	unsigned int	unused:7;

} s_demonio;

typedef struct _s_head {
	thinc_list 	list;
	int		sock;
	s_demonio*	demonio;
	/* function pointers */
	int 		(*read)(struct _s_head*, void*, size_t);
	int		(*write)(struct _s_head*, const void*, size_t);
} s_head;

#define s_head(p) ((p)->head)
#define s_headp(p) (&(p)->head)
#define s_sock(p) (s_head(p).sock)
#define s_list(p) (&(s_head(p).list))
#define s_dem(p)  (s_head(p).demonio)
#define s_readf(p) (s_head(p).read)
#define s_writef(p) (s_head(p).write)

typedef struct {
	s_head 	head;
	SSL*	ssl;
	char*	write_buf;
} s_client;

#define S_BUF_SIZE	1024

typedef struct {
	s_head head;
	char*	data;
	char*	shm_buf;
	int	shm_id;
	unsigned int shm_size;
} s_app;

/* 128K buffer for sound data */
#define S_SHM_BUF_SIZE	(1U<<17)

#define all_clients(cl, d, f, p) do { 					\
	s_client* __tmp; 						\
	t_list_for_each_entry(cl, __tmp, s_client, &d->clients, head.list)\
	{ 								\
		/* function call */					\
		err = f p;						\
		if(err == T_ERROR) {					\
			s_err("cannot call the function for the client\n");\
			s_client_die(d, cl);				\
		}							\
	}								\
} while(0)

/* XXX HACK XXX 
 * for SSL support. This should be fixed throughout THINC.. 
 */
const char *cert_path = "/etc/X11/thinc/certs/cacert.pem";
const char *key_path = "/etc/X11/thinc/certs/privkey.pem";

/* need this guy global for signal handler */
static s_demonio el_demonio;

static void usage(const char *name);
static int s_init_sockets(s_demonio *demonio, int port);
static int s_socket_unix(s_demonio *demonio);
static int s_socket_net(s_demonio *demonio, int port);
static void s_exit_sockets(s_demonio *demonio);
static inline int s_set_noblock(int sock);
static void die_die_die(int signal);
static void s_exit(s_demonio *demonio);
static int s_app_new(s_demonio *demonio);
static void s_app_die(s_demonio *demonio, s_app *app);
static int s_app_do(s_demonio *demonio, s_app *app);
static int s_client_new(s_demonio *demonio);
static void s_client_die(s_demonio *demonio, s_client *client);
static int s_client_do(s_demonio *demonio, s_client *client);
static int s_read(s_head* head, void *buf, size_t size);
static int s_write(s_head *head, const void *buf, size_t size);
static int s_write_sec(s_head *head, const void *buf, size_t size);
static int s_read_sec(s_head *head, void *buf, size_t size);
static int s_do_poll_out(int sock);

static int s_send_open(s_client *client, struct thinc_snd_cmd_open *cmd);
static int s_play(s_app *app, struct thinc_snd_cmd_play *cmd);
static int s_send_play(s_client* client, struct thinc_snd_cmd_play *cmd, 
	char *buf);

int main(int argc, char **argv)
{
	int port = T_SND_PORT;
	int opt, index = 0;
	fd_set read_fds;
	struct option long_options[] = {
		{"help", 0, 0, 'h'},
		{"port", 1, 0, 'p'},
		{"no-ssl", 0, 0, 's'},
		{"verbose", 0, 0, 'v'},
		{0, 0, 0, 0}	/* last row must be filled with zeroes */
	};
	memset(&el_demonio, 0, sizeof(el_demonio));
	el_demonio.secure = 1;

	while(1) {
		opt = getopt_long(argc, argv, "hp:sv", long_options, &index);
		if (opt == -1) {
			/* finished */
			break;
		}

		switch (opt) {
			case 'h': 
				usage(argv[0]);
				exit(EXIT_SUCCESS);
			case 'v':
				verbose = 1;
				s_verb("Being verbose\n");
				break;
			case 'p':
				port = (int)strtol(optarg, (char **)NULL, 10);
				if ((port == EINVAL) || (port == ERANGE) ||
				    (port > (1<<16))
				) {
					s_err("Invalid port %d specified\n", 
						port);
					exit(EXIT_FAILURE);
				}
				s_verb("Using port %d.\n", port);
				break;
			case 's':
				s_warn("Disabling encryption.\n");
				el_demonio.secure = 0;
				break;
			default:
				usage(argv[0]);
				exit(EXIT_FAILURE);
		}
	}

	if(s_init_sockets(&el_demonio, port) == T_ERROR) {
		s_err("couldn't initialize sockets\n");
		exit(EXIT_FAILURE);
	}

	t_list_init(&el_demonio.apps);
	t_list_init(&el_demonio.clients);

	signal(SIGTERM, die_die_die);
	signal(SIGINT, die_die_die);
	
	do {
		int err;
		s_app *ca, *na;
		s_client *cc, *nc;
		s_verb("now waiting foreva for something to happen...\n");
again:
		memcpy(&read_fds, &el_demonio.fds, sizeof(read_fds));
		err = select(el_demonio.fdmax+1, &read_fds, NULL, 
			NULL, NULL);
		if(err == -1) {
			if(errno == EINTR)
				goto again;
			else {
				s_err("Couldn't select: %s\n", strerror(errno));
				break;
			}
		}
		s_dbg("something happened... in %d sockets\n", err);
		if(FD_ISSET(el_demonio.unix_fd, &read_fds)) {
			s_dbg("in the unix socket!\n");
			/* nothing to do on error */
			s_app_new(&el_demonio);
			--err;
			if(!err)
				continue;
		}
		if(FD_ISSET(el_demonio.net_fd, &read_fds)) {
			s_dbg("in the network socket!\n");
			/* nothing to do on error */
			s_client_new(&el_demonio);
			--err;
			if(!err)
				continue;
		}
		/* now chk the apps */
		t_list_for_each_entry(ca, na, s_app, &el_demonio.apps, 
			head.list) 
		{
			s_dbg("chking app %p\n", ca);
			if(FD_ISSET(s_sock(ca), &read_fds)) {
				s_dbg("app %p is alive!\n", ca);
				if(s_app_do(&el_demonio, ca) != T_OK) {
					s_app_die(&el_demonio, ca);
				}
				--err;
				if(!err)
					continue;
			}
		}
		/* now chk the clients */
		t_list_for_each_entry(cc, nc, s_client, &el_demonio.clients, 
			head.list) 
		{
			if(FD_ISSET(s_sock(cc), &read_fds)) {
				s_dbg("client %p is alive!\n", cc);
				if(s_client_do(&el_demonio, cc) != T_OK) {
					s_client_die(&el_demonio, cc);
				}
				--err;
				if(!err)
					continue;
			}
		}
		/* paranoia */
		if(err) {
			s_err("didn't process all sockets. %d left\n", err);
			break;
		}
	} while(1);

	s_exit(&el_demonio);

	s_msg("bye.\n");
	return EXIT_SUCCESS;
}

static void s_exit(s_demonio *demonio)
{
	s_exit_sockets(demonio);
}

static void die_die_die(int signal)
{
	s_exit(&el_demonio);
#if 0
	s_msg("death, at last\n");
#else
	s_msg("Et tu, Brute...\n");
#endif
	exit(EXIT_FAILURE);
}

static void usage(const char *name)
{
	printf("Usage: %s [options]\n", name);
	printf( "\nOptions:\n"
		"\t-h/--help          Print this message and exit\n"
		"\t-p/--port          Port to listen on (default %d\n"
		"\t-S/--no-ssl        Disable ssl\n"
		, T_SND_PORT
	);
}

static int s_init_sockets(s_demonio *demonio, int port)
{
	s_verb("opening local socket\n");

	if(s_socket_unix(demonio) == T_ERROR)
		return T_ERROR;

	s_verb("opening network socket\n");

	if(s_socket_net(demonio, port) == T_ERROR)
		return T_ERROR;

	FD_ZERO(&demonio->fds);
	FD_SET(demonio->unix_fd, &demonio->fds);
	FD_SET(demonio->net_fd, &demonio->fds);
	demonio->fdmax = (demonio->unix_fd > demonio->net_fd ? 
		demonio->unix_fd : 
		demonio->net_fd);

	s_dbg("finished\n");
	return 0;
}

static void s_exit_sockets(s_demonio *demonio)
{
	close(demonio->unix_fd);
	close(demonio->net_fd);
	unlink(demonio->un_path);
	free(demonio->un_path);
}

static int s_socket_unix(s_demonio *demonio)
{
	int s_len;
	struct sockaddr_un addr;
	socklen_t addr_len;
	int fd;

	s_dbg("opening new unix socket\n");

	/* i'm lazy, let's just try to create the dir */
	if(mkdir(SOCKET_DIR, 0700) == -1) {
		if(errno == EEXIST) {
			s_dbg("socket dir exists, chking permissions\n");
			if(chmod(SOCKET_DIR, 0700) == -1) {
				s_err("couldn't set up correct permissions "
					"for socket dir %s: %s\n", SOCKET_DIR,
					strerror(errno));
				return T_ERROR;
			}
		} else {
			s_err("couldn't create socket dir %s: %s\n", 
				SOCKET_DIR, strerror(errno));
			return T_ERROR;
		}
	}

	/* path + uid (max 64K?) + \0 */
	s_len = strlen(SOCKET_PATH);
	demonio->un_path_len = s_len + 6;
	demonio->un_path = calloc(1, demonio->un_path_len);
	if(!demonio->un_path) {
		s_err("you have no memory: %s\n", strerror(errno));
		return T_ERROR;
	}
	strncpy(demonio->un_path, SOCKET_PATH, s_len);
	snprintf(demonio->un_path+s_len, 5, "%u", getuid());

	if(demonio->un_path_len >= sizeof(addr.sun_path)) {
		s_err("socket pathane is too long (%d), max is %d\n",
			demonio->un_path_len, sizeof(addr.sun_path));
		return T_ERROR;
	}

	fd = socket(PF_UNIX, SOCK_STREAM, 0);
	if(fd == -1) {
		s_err("can't create local socket: %s\n", strerror(errno));
		return T_ERROR;
	}

	if(s_set_noblock(fd) == T_ERROR) {
		s_err("can't set local socket non-blocking: %s\n",
			strerror(errno));
		close(fd);
		return T_ERROR;
	}
	memset(&addr, 0, sizeof(addr));
	addr.sun_family = AF_LOCAL;
	strncpy(addr.sun_path, demonio->un_path, demonio->un_path_len);

	s_dbg("socket name %s\n", addr.sun_path);
	addr_len = sizeof(addr);

	if(bind(fd, (struct sockaddr *)&addr, addr_len) == -1) {
		if(errno == EADDRINUSE)
			s_err("another daemon is already running for this "
				"user (uid: %d). if this is not the case "
				"remove %s and try again\n", getuid(),
				demonio->un_path);
		else 
			s_err("can't bind socket: %s\n", strerror(errno));
		close(fd);
		return T_ERROR;
	}

	if(listen(fd, 2) == -1) {
		s_err("can't listen on socket: %s\n", strerror(errno));
		unlink(demonio->un_path);
		close(fd);
		return T_ERROR;
	}

	s_dbg("setting mode for socket\n");
	if(chmod(demonio->un_path, 0700) == -1) {
		s_err("couldn't set up correct permissions for socket %s: "
			"%s\n", demonio->un_path, strerror(errno));
		unlink(demonio->un_path);
		close(fd);
		return T_ERROR;
	}

	demonio->unix_fd = fd;

	s_dbg("done.\n");
	return T_OK;
}

static int s_socket_net(s_demonio *demonio, int port)
{
	struct sockaddr_in addr_s;
	socklen_t addrlen;
	int sock, reuse;

	s_dbg("initializing network socket\n");

	if ((sock = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
		s_err("Couldn't open network socket: %s\n", strerror(errno));
		return T_ERROR;
	}

	reuse = 1;
	if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &reuse, 
		sizeof(reuse)) == -1) 
	{
		s_err("Couldn't set up network socket: %s\n", strerror(errno));
		close(sock);
		return T_ERROR;
	}

#if 0
	if(s_set_noblock(sock) == T_ERROR) {
		s_err("Can't set local socket non-blocking: %s\n",
			strerror(errno));
		close(sock);
		return T_ERROR;
	}
#endif

	addr_s.sin_family = AF_INET;
	addr_s.sin_port = htons(port);
	addr_s.sin_addr.s_addr = htonl(INADDR_ANY);
	addrlen = sizeof(addr_s);

	if(bind(sock, (struct sockaddr*) &addr_s, addrlen) == -1) {
		s_err("Couldn't bind network socket: %s\n",
			strerror(errno));
		close(sock);
		return T_ERROR;
	}

	/* We don't anticipate much backlog for connection queues, which is what
	 * the second argument for listen() is specifying. */
	if(listen(sock, 1) == -1 ) {
		s_err("Couldn't set up network socket for listening: %s\n",
			strerror(errno));
		close(sock);
		return T_ERROR;
	}

	s_verb("listening on port %d\n", port);

	/* check for SSL */
	if(demonio->secure) {
		demonio->ctx = thinc_ssl_init_server(cert_path, key_path);
		if (!demonio->ctx) {
			s_err("Error starting SSL.\nDid you forget "
				"to put the certificates in %s and %s?\n"
				"Or, did you set the permissions "
				"for these certificates to g+r,o+r?\n", 
				cert_path, key_path);
			close(sock);
			return T_ERROR;
		}
		s_msg("SSL enabled\n");
	}

	demonio->net_fd = sock;

	s_dbg("done.\n");
	return T_OK;
}

static inline int s_set_noblock(int sock)
{
	int opts;

	opts = fcntl(sock,F_GETFL);
	if (opts < 0) {
		perror("fcntl(F_GETFL)");
		return T_ERROR;
	}
	opts = (opts | O_NONBLOCK);
	if (fcntl(sock,F_SETFL,opts) < 0) {
		return T_ERROR;
	}
	return T_OK;
}

/* called when get a new connection on the unix socket */
static int s_app_new(s_demonio *demonio)
{
	int sock;
	s_app *app;
	struct thinc_snd_cmd cmd;

	s_dbg("got new connection on unix socket\n");

	sock = accept(demonio->unix_fd, NULL, 0);
	if(sock == -1) {
		s_err("couldn't accept new application: %s\n",
			strerror(errno));
		return T_ERROR;
	}

#if 0
	if(s_set_noblock(sock) == T_ERROR) {
		s_err("Can't set new app socket non-blocking: %s\n",
			strerror(errno));
		close(sock);
		return T_ERROR;
	}
#endif

	app = calloc(1, sizeof(*app));
	if(!app) {
		s_err("No memory for new app :'(: %s\n", strerror(errno));
		close(sock);
		return T_ERROR;
	}

	app->shm_size = S_SHM_BUF_SIZE;
	app->shm_id = shmget(IPC_PRIVATE, app->shm_size,
		(SHM_R | SHM_W | 0600));
	if(app->shm_id == -1) {
		s_err("can't create shared memory for new app: %s\n",
			strerror(errno));
		free(app);
		close(sock);
		return T_ERROR;
	}
	app->shm_buf = shmat(app->shm_id, 0, 0);
	if(!app->shm_buf) {
		shmctl(app->shm_id, IPC_RMID, 0);
		free(app);
		close(sock);
	}
	app->data = app->shm_buf;
	
	s_sock(app) = sock;
	t_list_init(s_list(app));
	t_list_add_head(s_list(app), &demonio->apps);

	FD_SET(s_sock(app), &demonio->fds);
	if(s_sock(app) > demonio->fdmax)
		demonio->fdmax = s_sock(app);
	s_dem(app) = demonio;

	s_readf(app) = s_read;
	s_writef(app) = s_write;

	/* now send hello message to plugin with needed info */
	cmd.id = T_SND_CMD_INIT;
	cmd.type.init.shm_id = app->shm_id;
	cmd.type.init.shm_buf_sz = app->shm_size;
	if(s_writef(app)(s_headp(app), &cmd, sizeof(cmd)) == T_ERROR) {
		s_err("cannot send INIT cmd to plugin\n");
		s_app_die(demonio, app);
		return T_ERROR;
	}

	s_msg("Info: got new app connection\n");
	return T_OK;
}

static void s_app_die(s_demonio *demonio, s_app *app)
{
	/* we could try to fix up demonio->fdmax but seems too much pain */
	close(s_sock(app));
	FD_CLR(s_sock(app), &demonio->fds);
	t_list_del_el(s_list(app));
	shmdt(app->shm_buf);
	shmctl(app->shm_id, IPC_RMID, 0);
	free(app);
}

static int s_app_do(s_demonio *demonio, s_app *app)
{
	int err;
	struct thinc_snd_cmd cmd;
	s_client *cl;

	err = s_readf(app)(s_headp(app), &cmd, sizeof(cmd));
	if(err == T_WOULDBLOCK) {
		/* TODO */
		s_err("WOULDBLOCK!\n");
		return T_ERROR;
	} else if(err == T_ERROR) {
		return T_ERROR;
	}

	switch(cmd.id) {
		case T_SND_CMD_OPEN:
			s_dbg("got open\n");
			all_clients(cl, s_dem(app), s_send_open, 
				(cl, &cmd.type.open));
			break;
		case T_SND_CMD_PLAY:
			s_dbg("got play\n");
			err = s_play(app, &cmd.type.play);
			if(err == T_ERROR) {
				s_err("cannot issue play command. death\n");
				return T_ERROR;
			}
			break;
		case T_SND_CMD_CLOSE:
			s_dbg("got close\n");
	}

	s_dbg("done\n");
	return T_OK;
}

static int s_send_open(s_client *client, struct thinc_snd_cmd_open *cmd)
{
	int err;

	thinc_header *h;
	thinc_sndOpen *open;

	h = (thinc_header*) client->write_buf;
	h->type = T_SND_OPEN;
	open = (thinc_sndOpen*) (client->write_buf + THINC_HEADERSIZE);
	open->rate = htons(cmd->rate);
	open->bits = cmd->bits;
	open->channels = cmd->channels;
	open->endian = cmd->endian;

	s_dbg("sending open command. bits %d, channels %d, rate %d\n",
		open->bits, open->channels, cmd->rate);

	err = s_writef(client)(s_headp(client), client->write_buf, 
		THINC_HEADERSIZE + T_SND_OPEN_SIZE);
	if(err == T_ERROR) {
		s_err("can't write open header %d\n", err);
		return err;
	}

	s_dbg("finished\n");
	return T_OK;
}

static int s_play(s_app *app, struct thinc_snd_cmd_play *cmd)
{
	s_client *cl;
	int err;

	s_dbg("playing %u bytes of audio data\n", cmd->length);
	if(cmd->offset == 0) {
		/* rewind time */
		app->data = app->shm_buf;
	}
#if 1
	/* paranoia? */
	if(cmd->offset != app->data - app->shm_buf) {
		s_warn("offsets not in sync with app: %d vs %d\n", cmd->offset,
			app->data - app->shm_buf);
	}
#endif

	all_clients(cl, s_dem(app), s_send_play, (cl, cmd, app->data));

	app->data += cmd->length;

	s_dbg("finished\n");
	return T_OK;
}

static int s_send_play(s_client* client, struct thinc_snd_cmd_play *cmd, 
	char *buf)
{
	int err;
	thinc_header *h;
	thinc_sndPlay *play;

	s_dbg("sending play command. %u bytes of data\n", 
		cmd->length);

	h = (thinc_header*) client->write_buf;
	h->type = T_SND_PLAY;
	play = (thinc_sndPlay*) (client->write_buf + THINC_HEADERSIZE);
	play->timestamp = htonl(cmd->timestamp);
	/* just set the ID to 0 now. When we support mixing, this 
	 * will become significant. */
	play->id = 0;
	play->size = htonl(cmd->length);

	err = s_writef(client)(s_headp(client), client->write_buf, 
		THINC_HEADERSIZE + T_SND_PLAY_SIZE);
	if(err == T_ERROR) {
		s_err("can't write open header %d\n", err);
		return err;
	}

	err = s_writef(client)(s_headp(client), buf, cmd->length);
	if(err == T_ERROR) {
		s_err("can't write open header %d\n", err);
		return err;
	}

	s_dbg("finished\n");
	return T_OK;
}

/* called when get a new connection on the network socket */
static int s_client_new(s_demonio *demonio)
{
	int sock, err;
	s_client *client;

	s_dbg("got new connection on network socket\n");

	sock = accept(demonio->net_fd, NULL, 0);
	if(sock == -1) {
		s_err("couldn't accept new client: %s\n",
			strerror(errno));
		return T_ERROR;
	}

	client = calloc(1, sizeof(*client));
	if(!client) {
		s_err("No memory for new client :'(: %s\n", strerror(errno));
		close(sock);
		return T_ERROR;
	}
	
	s_sock(client) = sock;
	if(demonio->secure) {
		s_dbg("initializing ssl connection\n");
		err = thinc_ssl_accept(s_sock(client), demonio->ctx, 
			&client->ssl);
		if (err == T_SSL_CLOSED || err == T_SSL_ERROR ) {
			s_err("Couldn't initialize secure connection with new "
				"client: %d\n", err);
			close(sock);
			free(client);
			return T_ERROR;
		}
		s_readf(client) = s_read_sec;
		s_writef(client) = s_write_sec;
		s_dbg("ssl connected, wee\n");
	} else {
		s_warn("using insecure client connection\n");
		s_readf(client) = s_read;
		s_writef(client) = s_write;
	}

	/* FIXME: after ssl so don't have to deal with ssl handshake not 
	 * finishing */
#if 0
	if(s_set_noblock(sock) == T_ERROR) {
		s_err("Can't set new client socket non-blocking: %s\n",
			strerror(errno));
		close(sock);
		return T_ERROR;
	}
#endif

	client->write_buf = calloc(1, S_BUF_SIZE);
	if(!client->write_buf) {
		s_err("No memory for new client's buffer :'(: %s\n", 
			strerror(errno));
		free(client);
		close(sock);
		return T_ERROR;
	}

	t_list_init(s_list(client));
	t_list_add_head(s_list(client), &demonio->clients);

	FD_SET(s_sock(client), &demonio->fds);
	if(s_sock(client) > demonio->fdmax)
		demonio->fdmax = s_sock(client);
	s_dem(client) = demonio;

	s_msg("Info: got new client connection\n");
	return T_OK;
}

static void s_client_die(s_demonio *demonio, s_client *client)
{
	s_err("client dying :'(\n");
	/* we could try to fix up demonio->fdmax but seems too much pain */
	close(s_sock(client));
	FD_CLR(s_sock(client), &demonio->fds);
	t_list_del_el(s_list(client));
	free(client->write_buf);
	free(client);
}

static int s_client_do(s_demonio *demonio, s_client *client)
{
	int err;
	char a;

	do {
		err = s_readf(client)(s_headp(client), &a, 1);
		if(err == T_OK)
			s_dbg("got %c\n", a);
		else if(err == T_WOULDBLOCK)
			break;
		else 
			return T_ERROR;
	} while(1);

	return T_OK;
}

static int s_read(s_head *head, void *buf, size_t size)
{
	int err;
again:
	err = read(head->sock, buf, size);
	if(err == 0) {
		s_msg("Info: thang gone, bye bye\n");
		return T_ERROR;
	} else if(err == -1) {
		if(errno == EINTR) {
			goto again;
		} else if(errno == EWOULDBLOCK || errno == EAGAIN) {
			return T_WOULDBLOCK;
		} else {
			s_err("can't read: %s\n", strerror(errno));
			return T_ERROR;
		}
	}

	return err;
}

static int s_write(s_head *head, const void *buf, size_t size)
{
	int err;

	do {
		err = write(head->sock, buf, size);
		if(err == -1) {
			if(errno == EINTR) {
				continue;
			} else if (errno == EWOULDBLOCK || errno == EAGAIN) {
				if(s_do_poll_out(head->sock) == T_ERROR) {
					s_err("got tired of waiting\n");
					return T_ERROR;
				}
				continue;
			} else {
				s_err("can't write: %s\n", strerror(errno));
				return T_ERROR;
			}
		}
		size -= err;
		buf += err;
	} while(size > 0);

	return size;
}

static int s_read_sec(s_head *head, void *buf, size_t size)
{
	int err;
	/* apps not supposed to use this function so this is ok */
	s_client *cl = (s_client*)head;
	err = thinc_ssl_read(cl->ssl, buf, size);
	if(err == T_SSL_CLOSED) {
		s_msg("Info: thang gone, bye bye\n");
		return T_ERROR;
	} else if(err == T_SSL_WOULDBLOCK) {
		return T_WOULDBLOCK;
	} else if(err == T_SSL_ERROR) {
		s_err("can't ssl read\n");
		return T_ERROR;
	}

	return err;
}

static int s_write_sec(s_head *head, const void *buf, size_t size)
{
	int err;
	/* apps not supposed to use this function so this is ok */
	s_client *cl = (s_client*)head;

	do {
		err = thinc_ssl_write(cl->ssl, buf, size);
		if(err == T_SSL_CLOSED) {
			s_msg("Info: thang gone, bye bye\n");
			return T_ERROR;
		} else if(err == T_SSL_WOULDBLOCK) {
			if(s_do_poll_out(s_sock(cl)) == T_ERROR)
				return T_ERROR;
			continue;
		} else if(err == T_SSL_ERROR) {
			s_err("can't ssl write\n");
			return T_ERROR;
		}
		size -= err;
		buf += err;
	} while(size > 0);

	return size;
}

static int s_do_poll_out(int sock)
{
	struct pollfd fds[1];
	int err;

	fds[0].fd = sock;
	fds[0].events = POLLOUT;
	fds[0].revents = 0;

again:
	/* give it a max of 100ms to respond */
	err = poll(fds, 1, 500);
	if(err == -1) {
		if(errno == EINTR)      /* try again */
			goto again;
		return T_ERROR;
	} else if(err == 0) { /* timeout */
		s_err("timeout while trying to send data\n");
		return T_ERROR;
	} else if(fds[0].revents & (POLLERR | POLLHUP | POLLNVAL)) {
		return T_ERROR;
	}

	return T_OK;
}
