/* Copyright 2007 Ricardo Baratto */
/* Copyright 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * THINC alsa plugin
 *
 * Authors: 
 * Leo Kim - Network Computing Lab - Columbia University
 * Ricardo Baratto - Network Computing Lab - Columbia University
 */
/* set tabstop = 8 */

/* for full details, see:
 * http://www.alsa-project.org/alsa-doc/alsa-lib/pcm_external_plugins.html
 */

#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/un.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>

#include <alsa/asoundlib.h>
#include <alsa/pcm_external.h>

#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/shm.h>

#include "common.h"

/* Data types */
typedef struct snd_pcm_thinc {
	snd_pcm_ioplug_t io;
	int fd;
	unsigned int frame_size;
	unsigned int hw_ptr;
	unsigned int rate;
	int shm_id;
	char *shm_buff;
	char *shm_end;
	char *data;
} snd_pcm_thinc_t;

/* Macros */
#define ARRAY_NELEMS(a) (sizeof(a) / sizeof(a[0]) )

/* Function declarations */
static snd_pcm_sframes_t thinc_pcm_write(snd_pcm_ioplug_t *io,
	const snd_pcm_channel_area_t *areas, snd_pcm_uframes_t offset,
	snd_pcm_uframes_t size);
#if 0 /* for capture */
static snd_pcm_sframes_t thinc_pcm_read(snd_pcm_ioplug_t *io,
	const snd_pcm_channel_area_t *areas, snd_pcm_uframes_t offset,
	snd_pcm_uframes_t size);
#endif
static snd_pcm_sframes_t thinc_pcm_pointer(snd_pcm_ioplug_t *io);
static int thinc_pcm_start(snd_pcm_ioplug_t *io);
static int thinc_pcm_stop(snd_pcm_ioplug_t *io);
static int thinc_pcm_prepare(snd_pcm_ioplug_t *io);
static int thinc_pcm_hw_params(snd_pcm_ioplug_t *io, 
	snd_pcm_hw_params_t *params);
static int thinc_pcm_hw_constraint(snd_pcm_thinc_t *spt);
static int thinc_pcm_close(snd_pcm_ioplug_t *io);

static int thinc_pcm_connect(snd_pcm_thinc_t *spt);
static int thinc_pcm_send(snd_pcm_thinc_t *spt, void *buf, size_t len);
static int thinc_pcm_read(snd_pcm_thinc_t *spt, void *buf, size_t len);

/* struct definitions. this is what alsa will come looking for */
static snd_pcm_ioplug_callback_t thinc_pcm_playback_callback = {
	/* mandatory */
	.start = thinc_pcm_start,
	.stop = thinc_pcm_stop,
	.pointer = thinc_pcm_pointer,
	/* optional */
	.close = thinc_pcm_close,
	.hw_params = thinc_pcm_hw_params,
	.prepare = thinc_pcm_prepare,
	.transfer = thinc_pcm_write,
};

/* This is the struct with callbacks that handle sound capture */
#if 0
static snd_pcm_ioplug_callback_t thinc_pcm_capture_callback = {
	.close = thinc_pcm_close,
	.hw_params = thinc_pcm_hw_params,
	.start = thinc_pcm_start,
	.stop = thinc_pcm_stop,
	.prepare = thinc_pcm_prepare,
	.transfer = thinc_pcm_read,
	.pointer = thinc_pcm_pointer,
};
#endif

/************************************************************************************
 * Alsa Plugin Functions
 ***********************************************************************************/

/* this one's name is _snd_pcm_thinc_open. this is where initialization should
 * happen (basically open socket to tsd) */
SND_PCM_PLUGIN_DEFINE_FUNC(thinc)
{
#if 0
	/* TODO: for configuration options */
	snd_config_iterator_t i, next;
#endif
	int err;
	snd_pcm_thinc_t *spt = 0;

	s_dbg("Open mode is for %s\n", stream == SND_PCM_STREAM_PLAYBACK ? 
		"Playback" : "Capture");

	/* Here's where you initialize the "io" struct stuff. Check the ALSA
	 * doxygen docs for details */
	spt = calloc(1, sizeof(snd_pcm_thinc_t));
	if(!spt)
		return -ENOMEM;
	spt->io.name = "THINC Audio";
	/* poll_events, mmap_rw, poll_fd are kinda funky. from the docs:
	 * - poll_fd and poll_events specify the poll file descriptor and the
	 *   corresponding poll events (POLLIN, POLLOUT) for the plugin. If the
	 *   plugin requires multiple poll descriptors or poll descriptor(s)
	 *   dynamically varying, set poll_descriptors and
	 *   poll_descriptors_count callbacks to the callback table. Then the
	 *   poll_fd and poll_events field are ignored.
	 * - mmap_rw specifies whether the plugin behaves in the pseudo mmap
	 *   mode. When this value is set to 1, the plugin creates always a
	 *   local buffer and performs read/write calls using this buffer as if
	 *   it's mmapped. The address of local buffer can be obtained via
	 *   snd_pcm_ioplug_mmap_areas() function.
	 */

	spt->io.poll_events = POLLOUT;
	spt->io.mmap_rw	= 0;
	spt->io.poll_fd	= -1;
	spt->io.version	= SND_PCM_IOPLUG_VERSION;
	spt->io.callback = &thinc_pcm_playback_callback;

	if(thinc_pcm_connect(spt) == T_ERROR) {
		free(spt);
		return -1;
	}

	if ( (err = snd_pcm_ioplug_create(&spt->io, name, stream, mode)) < 0) {
		s_err("Oh shit, couldn't create ioplug struct!");
		free(spt);
		return err;
	}

	if ( (err = thinc_pcm_hw_constraint(spt)) < 0) {
		s_err("Oh shit, an ERROR HAS OCCURRED!\n");
		snd_pcm_ioplug_delete(&spt->io);
		return err;
	}

	spt->io.private_data = spt;
	*pcmp = spt->io.pcm;

	s_dbg("Opened as ioplug=%p, pcm=%p, ioplug->callback = %p\n", &spt->io,
		spt->io.pcm, spt->io.callback);

	s_dbg("Finished.\n");
	return 0;

#if 0
	/* Here's how you can parse the asound.conf file and get funky values */
	snd_config_for_each(i, next, conf) {
		snd_config_t *n = snd_config_iterator_entry(i);
		const char *id;
		if (snd_config_get_id(n, &id) < 0)
			continue;
		if (strcmp(id, "comment") == 0 || strcmp(id, "type") == 0)
			continue;
		if (!strcmp(id, "bdaddr")) {
			const char *addr;
			if (snd_config_get_string(n, &addr) < 0) {
				SNDERR("Invalid type for %s", id);
				return -EINVAL;
			}
			str2ba(addr, &hs_bdaddr);
			continue;
		}
		if (!strcmp(id, "timeout")) {
			if (snd_config_get_integer(n, &timeout) < 0) {
				SNDERR("Invalid type for %s", id);
				return -EINVAL;
			}
			continue;
		}
		SNDERR("Unknown field %s", id);
		return -EINVAL;
	}
#endif
}

SND_PCM_PLUGIN_SYMBOL(thinc);

/* called (by us) to set the hardware constraints. kind of random/copy for now as
 * this is all kind of obscure */
static int thinc_pcm_hw_constraint(snd_pcm_thinc_t *spt)
{	
	s_dbg("thinc_pcm_hw_constraint.\n");
	snd_pcm_ioplug_t *io = &spt->io;
	int err;
	static const snd_pcm_access_t access_list[] = {
		SND_PCM_ACCESS_RW_INTERLEAVED,
		SND_PCM_ACCESS_MMAP_INTERLEAVED
	};
	/* leok: We make assumptions about the formats we support. Probably not
	 * such a good idea since the client might not support them. */
	static const unsigned int format[] = {
		SND_PCM_FORMAT_S16_LE,
		SND_PCM_FORMAT_U8,
		SND_PCM_FORMAT_S16_BE,
		SND_PCM_FORMAT_MU_LAW,
		SND_PCM_FORMAT_S16,
	};

	/* Access type */
	err = snd_pcm_ioplug_set_param_list(io, SND_PCM_IOPLUG_HW_ACCESS,
			ARRAY_NELEMS(access_list), access_list);
	if(err < 0) {
		s_err("SND_PCM_IOPLUG_HW_ACCESS: %d\n", err);
		return err;
	}

	/* Supported formats */
	err = snd_pcm_ioplug_set_param_list(io, SND_PCM_IOPLUG_HW_FORMAT,
			ARRAY_NELEMS(format), format);
	if(err < 0) {
		s_err("SND_PCM_IOPLUG_HW_FORMAT: %d\n", err);
		return err;
	}

	/* Supported channels */
	/* leok: Assume 1- or 2-channel sound */
	err = snd_pcm_ioplug_set_param_minmax(io, SND_PCM_IOPLUG_HW_CHANNELS, 
			1, 2);
	if(err < 0) {
		s_err("SND_PCM_IOPLUG_HW_CHANNELS: %d\n", err);
		return err;
	}

	/* Period stuff */
	err = snd_pcm_ioplug_set_param_minmax(io, 
		SND_PCM_IOPLUG_HW_PERIOD_BYTES, 1, 4294967295U);
	if(err < 0) {
		s_err("SND_PCM_IOPLUG_HW_PERIOD_BYTES: %d\n", err);
		return err;
	}
	err = snd_pcm_ioplug_set_param_minmax(io, SND_PCM_IOPLUG_HW_PERIODS,
		2, 4294967295U);
	if(err < 0) {
		s_err("SND_PCM_IOPLUG_HW_PERIODS: %d\n", err);
		return err;
	}

	err = snd_pcm_ioplug_set_param_minmax(io, 
		SND_PCM_IOPLUG_HW_BUFFER_BYTES, 1, 4294967295U);
	if(err < 0) {
		s_err("SND_PCM_IOPLUG_HW_BUFFER_BYTES: %d\n", err);
		return err;
	}

	/* Supported rates */
	err = snd_pcm_ioplug_set_param_minmax(io, SND_PCM_IOPLUG_HW_RATE, 
			8000, 480000);
	if(err < 0) {
		s_err("SND_PCM_IOPLUG_HW_RATE: %d\n", err);
		return err;
	}

	s_dbg("finished\n");
	return 0;
}

/* mandatory: called when the PCM stream is started */
static int thinc_pcm_start(snd_pcm_ioplug_t *io)
{
	s_dbg("thinc_pcm_start\n");
	return 0;
}

/* mandatory: called when the PCM stream is stopped */
static int thinc_pcm_stop(snd_pcm_ioplug_t *io)
{
	s_dbg("thinc_pcm_stop\n");
	return 0;
}

/* mandatory: returns the current DMA position, which may be called at any 
 * time */
static snd_pcm_sframes_t thinc_pcm_pointer(snd_pcm_ioplug_t *io)
{
	snd_pcm_thinc_t *spt = io->private_data;
	s_dbg("ptr: %d\n", spt->hw_ptr);
	return spt->hw_ptr;
}

/* this is the transfer() callback. it's called when any data transfer happens.
 * It receives the area array, offset, and the size to transfer. The area array
 * contains the array of snd_pcm_channel_area_t with the elements of number of
 * channels.
 *
 * in thinc terms:
 * this function is responsible for determining how much data is written to
 * tsd. If tsd blocks, then this should return status to the calling audio
 * program appropriately with the number of frames written (which would be
 * zero if tsd isn't accepting any data). 
 */ 
static snd_pcm_sframes_t thinc_pcm_write(snd_pcm_ioplug_t *io,
	const snd_pcm_channel_area_t *areas, snd_pcm_uframes_t offset,
	snd_pcm_uframes_t size)
{
	snd_pcm_thinc_t *spt = io->private_data;
	struct thinc_snd_cmd cmd;
	char *data;
	unsigned long sleepy;

	s_dbg("sending size %lu, offset %lu, frame size %u\n", size, offset,
		spt->frame_size);

#if 0
	cmd.id = T_SND_CMD_PLAY;
	cmd.type.play.offset = 0; /* FIXME: legacy from kernel days */
	gettime_us(cmd.type.play.timestamp);
	cmd.type.play.length = size * spt->frame_size;

	s_dbg("sending %u bytes of audio data. timestamp %u\n", 
		cmd.type.play.length, cmd.type.play.timestamp);
	if(thinc_pcm_send(spt, &cmd, sizeof(cmd)) == T_ERROR) {
		s_err("can't send SND PLAY command to daemon\n");
		return -1;
	}

	s_dbg("getting data. first %d, step %d\n", areas->first, areas->step);
	data = (char *)areas->addr + 
		((areas->first + areas->step * offset) >> 3);
	if(thinc_pcm_send(spt, data, cmd.type.play.length) == T_ERROR) {
		s_err("can't send audio data to daemon\n");
		return -1;
	}
	spt->hw_ptr += size;
#else
	cmd.id = T_SND_CMD_PLAY;
	gettime_us(cmd.type.play.timestamp);
	cmd.type.play.length = size * spt->frame_size;
	if(spt->data + cmd.type.play.length < spt->shm_end) {
		/* still have space in the bluffer */
		cmd.type.play.offset = spt->data - spt->shm_buff;
	} else {
		/* no space, rewind */
		cmd.type.play.offset = 0;
		spt->data = spt->shm_buff;
		/* not cool to do this, causes I/O error. i guess it's just
		 * supposed to grow forever? */
		/*
		spt->hw_ptr = 0;
		*/
	}
	data = (char *)areas->addr + 
		((areas->first + areas->step * offset) >> 3);
	/* we copy the data first to make sure tsd reads the correct stuff */
	memcpy(spt->data, data, cmd.type.play.length);
	s_dbg("'sending' %u bytes of audio data. timestamp %u. our offset "
		"%d\n", cmd.type.play.length, cmd.type.play.timestamp,
		cmd.type.play.offset);
	if(thinc_pcm_send(spt, &cmd, sizeof(cmd)) == T_ERROR) {
		s_err("can't send SND PLAY command to daemon\n");
		return -1;
	}
	spt->data += cmd.type.play.length;
	spt->hw_ptr += size;
	sleepy = ((double)size/spt->rate)*1000000;
	sleepy -= (sleepy * 0.08);
	s_err("size %lu, sleeping for %lu\n", size, sleepy);

	usleep(sleepy);
#endif

	s_dbg("finished\n");
	return size;
}

/* called when snd_pcm_prepare() is called by an app: prepare PCM for use */
static int thinc_pcm_prepare(snd_pcm_ioplug_t *io)
{
	snd_pcm_thinc_t *spt = io->private_data;
	s_dbg("Resetting pointer to 0\n");
	spt->hw_ptr = 0;
	return 0;
}

#if 0 /* For capture */
static snd_pcm_sframes_t thinc_pcm_read(snd_pcm_ioplug_t *io,
	const snd_pcm_channel_area_t *areas, snd_pcm_uframes_t offset,
	snd_pcm_uframes_t size)
{
	s_dbg("thinc_pcm_read");
	return 0;
}
#endif

/* called when hw_params are set. Note that they may be called multiple times
 * according to the application */
static int thinc_pcm_hw_params(snd_pcm_ioplug_t *io, 
	snd_pcm_hw_params_t *params)
{
	struct thinc_snd_cmd cmd;
	snd_pcm_thinc_t *spt = io->private_data;

	spt->frame_size = (snd_pcm_format_physical_width(io->format) * 
		io->channels) >> 3;
	spt->rate = io->rate;
	s_dbg("setting hw params. frame size is %u\n", spt->frame_size);
	cmd.id = T_SND_CMD_OPEN;
	/* FIXME: how to figure this out? */
	cmd.type.open.bits = 16;
	cmd.type.open.channels = io->channels;
	cmd.type.open.rate = io->rate;
	/* FIXME: how to figure this out? */
	cmd.type.open.endian = T_LITTLE;

	s_dbg("sending open command: bits %d, channels %d, rate %d\n",
		cmd.type.open.bits, cmd.type.open.channels,
		cmd.type.open.rate);

	if(thinc_pcm_send(spt, &cmd, sizeof(cmd)) == T_ERROR) {
		s_err("can't send SND OPEN command to daemon\n");
		return -1;
	}

	s_dbg("finished\n");
	return 0;
}

/* When the PCM is closed, close callback is called. If the driver allocates
 * any internal buffers, they should be released in this callback.
 */
static int thinc_pcm_close(snd_pcm_ioplug_t *io)
{
	snd_pcm_thinc_t *spt = io->private_data;
	s_dbg("thinc_pcm_close\n");
	free(spt);
	io->private_data = 0;
	return 0;
}

/************************************************************************************
 * TSD Functions
 ***********************************************************************************/

/* connect to tsd */
static int thinc_pcm_connect(snd_pcm_thinc_t *spt)
{
	int name_len, s_len;
	struct sockaddr_un addr;
	socklen_t addr_len;
	struct thinc_snd_cmd cmd;
	int err;

	/* patch + uid (max 64K?) + \0 */
	s_len = strlen(SOCKET_PATH);
	name_len = s_len + 6;

	if(name_len >= sizeof(addr.sun_path)) {
		s_err("socket pathane is too long (%d, max is %d\n",
			name_len, sizeof(addr.sun_path));
		return T_ERROR;
	}

	spt->fd = socket(PF_UNIX, SOCK_STREAM, 0);
	if(spt->fd == -1) {
		s_err("can't create socket: %s\n", strerror(errno));
		return T_ERROR;
	}

	memset(&addr, 0, sizeof(addr));
	addr.sun_family = AF_LOCAL;
	strncpy(addr.sun_path, SOCKET_PATH, s_len);
	snprintf(addr.sun_path+s_len, 5, "%u", getuid());

	s_dbg("socket name %s\n", addr.sun_path);
	addr_len = sizeof(addr);

	if(connect(spt->fd, (struct sockaddr*) &addr, addr_len) == -1) {
		s_err("can't connect to tsd: %s\n", strerror(errno));
		close(spt->fd);
		return T_ERROR;
	}

	err = thinc_pcm_read(spt, &cmd, sizeof(cmd));
	if(err == 0 || err == T_ERROR) {
		s_err("can't read init message from tsd\n");
		close(spt->fd);
		return T_ERROR;
	}
	if(cmd.id != T_SND_CMD_INIT) {
		s_err("received unexpected command %d from tsd\n", cmd.id);
		close(spt->fd);
		return T_ERROR;
	}
	spt->shm_id = cmd.type.init.shm_id;
	spt->shm_buff = shmat(spt->shm_id, 0, 0);
	if(!spt->shm_buff) {
		s_err("can't attached shared memory: %s\n", strerror(errno));
		close(spt->fd);
		return T_ERROR;
	}
	spt->data = spt->shm_buff;
	spt->shm_end = spt->shm_buff + cmd.type.init.shm_buf_sz;

	s_dbg("finished\n");
	return T_OK;
}

/* send data to tsd */
static int thinc_pcm_send(snd_pcm_thinc_t *spt, void *buf, size_t len)
{
	int err;
again:
	err = write(spt->fd, buf, len);
	if(err == -1) {
		if(errno == EINTR) {
			goto again;
		} else if(errno == EWOULDBLOCK || errno == EAGAIN) {
#if 0
			return T_WOULDBLOCK;
#else
			s_err("WOULDBLOCK!\n");
			return T_ERROR;
#endif
		} else {
			s_err("can't write: %s\n", strerror(errno));
			return T_ERROR;
		}
	}

	return err;
}

/* read data from tsd */
static int thinc_pcm_read(snd_pcm_thinc_t *spt, void *buf, size_t len)
{
	int err;
again:
	err = read(spt->fd, buf, len);
	if(err == -1) {
		if(errno == EINTR) {
			goto again;
		} else if(errno == EWOULDBLOCK || errno == EAGAIN) {
#if 0
			return T_WOULDBLOCK;
#else
			s_err("WOULDBLOCK!\n");
			return T_ERROR;
#endif
		} else {
			s_err("can't read: %s\n", strerror(errno));
			return T_ERROR;
		}
	}

	return err;
}
